#pragma once

#include <folly/Format.h>
#include <folly/Optional.h>
#include <folly/Range.h>

#include <string>

namespace rtransfer {
namespace http {

namespace detail {
constexpr static const char* front =
    "GET /rtransfer HTTP/1.1\n"
    "Host: ";
constexpr static const char* rest =
    "\n"
    "Connection: upgrade\n"
    "Upgrade: clproto\n";
constexpr static const char* response =
    "101\n"
    "Connection: upgrade\n"
    "Upgrade: clproto\n";

constexpr std::size_t len(const char* str)
{
    return *str ? 1 + len(str + 1) : 0;
}
}  // namespace detail

inline std::string makeHeader(folly::StringPiece hostname)
{
    return detail::front + hostname.str() + detail::rest;
}

inline std::string makeResponse() { return detail::response; }

enum class ParseStatus { more_data, bad_header, ok };

inline std::pair<ParseStatus, std::size_t> parseHeader(
    folly::StringPiece header)
{
    if (header.size() < detail::len(detail::front))
        return {ParseStatus::more_data, 0};

    if (!header.startsWith(detail::front))
        return {ParseStatus::bad_header, 0};

    auto hostPiece = header.subpiece(detail::len(detail::front));
    const std::size_t hostEnd = hostPiece.find('\n');
    if (hostEnd == std::string::npos)
        return {ParseStatus::more_data, 0};

    if (hostEnd < 1 || hostEnd > 253)
        return {ParseStatus::bad_header, 0};

    auto restPiece = hostPiece.subpiece(hostEnd);
    if (restPiece.size() < detail::len(detail::rest))
        return {ParseStatus::more_data, 0};

    if (restPiece.removePrefix(detail::rest))
        return {ParseStatus::ok, restPiece.data() - header.data()};

    return {ParseStatus::bad_header, 0};
}

inline std::pair<ParseStatus, std::size_t> parseResponse(
    const folly::StringPiece resp)
{
    if (resp.size() < detail::len(detail::response))
        return {ParseStatus::more_data, 0};

    auto rest = resp;
    if (rest.removePrefix(detail::response))
        return {ParseStatus::ok, rest.data() - resp.data()};

    return {ParseStatus::bad_header, 0};
}

constexpr std::size_t responseLen() { return detail::len(detail::response); }

}  // namespace http
}  // namespace rtransfer
